//=============================================================================
// Lift exit type
// Targeted pathing to the lift center i'm aiming at
//
// For testing purposes, do not modify or redistribute
// Have suggestions? Contact me.
// caco_fff@hotmail.com - Higor at UT99.org
//=============================================================================
class Botz_LiftExit expands Botz_NavigBase;

const BFM = class'BotzFunctionManager';
var Mover MyMover;

//Called after botZ decides this is the path to take
//Do some lift handling here?
function bool PostPathEvaluate( botz other)
{
	if ( MyMover == none )
		return false;
		
	//Evaluate fall
	if ( BadFall(other) )
	{
		other.SpecialPause = 0.2;
		other.GotoState( other.GetStateName(), 'SpecialWait');
		return true;
	}
	
	if ( !other.PointReachable(location) )
	{
		//Temporary solution
		if ( (other.Base == MyMover) && !MyMover.bInterpolating  && !MyMover.bDelaying )
		{
			other.MoveTarget = other;
			other.SpecialPause = 0.5;
			other.bTickedJump = true;
		}
		else if ( other.Base == MyMover )
		{
			other.SpecialPause = 0.2;
			other.GotoState( other.GetStateName(), 'SpecialWait');
		}
		return true;
	}
	return false;
}


function Actor SpecialHandling(Pawn Other)
{

	if ( (Other.Base == MyMover) && (MyMover != None) )
	{
		if ( (self.Location.Z < Other.Location.Z + Other.CollisionHeight)
			 && Other.LineOfSightTo(self) )
			return self;
		Other.SpecialGoal = None;
		Other.DesiredRotation = rotator(Location - Other.Location);
		MyMover.HandleDoor(Other);

		if ( (Other.SpecialGoal == MyMover) || (Other.SpecialGoal == None) )
			Other.SpecialGoal = MyMover.myMarker;
		return Other.SpecialGoal;
	}
	return self;
}


function bool BadFall( Pawn Other)
{
	//Evaluate fall
	return (Other.Base == MyMover) && (MyMover.Velocity.Z < 0)
		&& !Region.Zone.bWaterZone
		&& (Location.Z < other.Location.Z)
		&& ((BFM.static.FreeFallVelocity( Location.Z-other.Location.Z, Region.Zone.ZoneGravity.Z) < -750 - other.JumpZ) //Fall too hard
			|| (BFM.static.FreeFallVelocity( Location.Z-other.Location.Z, Region.Zone.ZoneGravity.Z) > MyMover.Velocity.Z) ); //Fall slower than elevator
}


event FinishedPathing()
{
	local Botz_LiftCenter LC, best;
	local vector CmpPoint;

	ForEach AllActors (class'Botz_LiftCenter', LC)
	{
		CmpPoint = Location + vector(Rotation) * VSize(LC.Location - Location);
		if ( VSize(LC.Location - CmpPoint) < 180 )
		{
			if ( (best == none) || (VSize(Location - LC.Location) < VSize(Location - best.Location) ) )
				best = LC;
		}
	}

	if ( best != none )
	{
		if ( !bOneWayOut )
			AddPathHere( self, best, true);
		if ( !bOneWayInc )
			AddPathHere( best, self, true);
		MyMover = best.MyMover;
	}

}




defaultproperties
{
	FriendlyName="Lift Exit"
	MaxDistance=550
	Texture=Texture'BWP_Junction'
	bPushSave=True
	ReservePaths=1
	ReserveUpstreamPaths=1
}